# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
from typing import Tuple
from .models import Transaction, RefundReason
from .... import jsonrpc, identifier, offchain, stdlib, utils, txnmetadata, LocalAccount


@dataclass
class DiemAccount:
    account: LocalAccount
    client: jsonrpc.Client

    def refund_metadata(self, version: int, reason: RefundReason) -> Tuple[bytes, bytes]:
        return (txnmetadata.refund_metadata(version, reason.to_diem_type()), b"")

    def general_metadata(self, from_subaddress: bytes, payee: str) -> Tuple[bytes, bytes]:
        to_account, to_subaddress = identifier.decode_account(payee, self.account.hrp)
        return (txnmetadata.general_metadata(from_subaddress, to_subaddress), b"")

    def travel_metadata(self, cmd: offchain.PaymentCommand) -> Tuple[bytes, bytes]:
        metadata = cmd.travel_rule_metadata(self.account.hrp)
        return (metadata, bytes.fromhex(str(cmd.payment.recipient_signature)))

    def submit_p2p(self, txn: Transaction, metadata: Tuple[bytes, bytes]) -> str:
        to_account, to_subaddress = identifier.decode_account(str(txn.payee), self.account.hrp)
        script = stdlib.encode_peer_to_peer_with_metadata_script(
            currency=utils.currency_code(txn.currency),
            amount=txn.amount,
            payee=to_account,
            metadata=metadata[0],
            metadata_signature=metadata[1],
        )
        return self.account.submit_txn(self.client, script).bcs_serialize().hex()



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

"""Provides LocalAccount class for holding local account private key.

LocalAccount provides operations we need for creating auth key, account address and signing
raw transaction.
"""

from . import diem_types, jsonrpc, utils, stdlib, identifier
from .serde_types import uint64

from .auth_key import AuthKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey
from typing import Dict, Optional, Tuple, Union
from dataclasses import dataclass, field
from copy import copy
import time


@dataclass
class LocalAccount:
    """LocalAccount is like a wallet account

    Some default values are initialized for testnet.

    WARN: This is handy class for creating tests for your application, but may not ideal for your
    production code, because it uses a specific implementaion of ed25519 and requires loading your
    private key into memory and hand over to code from external.
    You should always choose more secure way to handle your private key
    (e.g. https://en.wikipedia.org/wiki/Hardware_security_module) in production and do not give
    your private key to any code from external if possible.
    """

    @staticmethod
    def generate() -> "LocalAccount":
        """Generate a random private key and initialize local account"""

        return LocalAccount()

    @staticmethod
    def from_private_key_hex(key: str) -> "LocalAccount":
        return LocalAccount.from_dict({"private_key": key})

    @staticmethod
    def from_dict(dic: Dict[str, str]) -> "LocalAccount":
        """from a dict that is created by LocalAccount#to_dict

        The private_key and compliance_key values are hex-encoded bytes; they will
        be loaded by `Ed25519PrivateKey.from_private_bytes`.
        """

        dic = copy(dic)
        for name in ["private_key", "compliance_key"]:
            if name not in dic:
                continue
            key = dic[name]
            dic[name] = Ed25519PrivateKey.from_private_bytes(bytes.fromhex(key))
        return LocalAccount(**dic)  # pyre-ignore

    private_key: Ed25519PrivateKey = field(default_factory=Ed25519PrivateKey.generate)
    compliance_key: Ed25519PrivateKey = field(default_factory=Ed25519PrivateKey.generate)

    hrp: str = field(default=identifier.TDM)
    txn_gas_currency_code: str = field(default="XDX")
    txn_max_gas_amount: int = field(default=1_000_000)
    txn_gas_unit_price: int = field(default=0)
    txn_expire_duration_secs: int = field(default=30)

    @property
    def auth_key(self) -> AuthKey:
        return AuthKey.from_public_key(self.public_key)

    @property
    def account_address(self) -> diem_types.AccountAddress:
        return self.auth_key.account_address()

    @property
    def public_key_bytes(self) -> bytes:
        return utils.public_key_bytes(self.public_key)

    @property
    def public_key(self) -> Ed25519PublicKey:
        return self.private_key.public_key()

    @property
    def compliance_public_key_bytes(self) -> bytes:
        return utils.public_key_bytes(self.compliance_key.public_key())

    def account_identifier(self, subaddress: Union[str, bytes, None] = None) -> str:
        return identifier.encode_account(self.account_address, subaddress, self.hrp)

    def decode_account_identifier(self, encoded_id: str) -> Tuple[diem_types.AccountAddress, Optional[bytes]]:
        return identifier.decode_account(encoded_id, self.hrp)

    def sign(self, txn: diem_types.RawTransaction) -> diem_types.SignedTransaction:
        """Create signed transaction for given raw transaction"""

        signature = self.private_key.sign(utils.raw_transaction_signing_msg(txn))
        return utils.create_signed_transaction(txn, self.public_key_bytes, signature)

    def create_txn(self, client: jsonrpc.Client, script: diem_types.Script) -> diem_types.SignedTransaction:
        sequence_number = client.get_account_sequence(self.account_address)
        chain_id = client.get_last_known_state().chain_id
        return self.sign(
            diem_types.RawTransaction(  # pyre-ignore
                sender=self.account_address,
                sequence_number=uint64(sequence_number),
                payload=diem_types.TransactionPayload__Script(value=script),
                max_gas_amount=uint64(self.txn_max_gas_amount),
                gas_unit_price=uint64(self.txn_gas_unit_price),
                gas_currency_code=self.txn_gas_currency_code,
                expiration_timestamp_secs=uint64(int(time.time()) + self.txn_expire_duration_secs),
                chain_id=diem_types.ChainId.from_int(chain_id),
            )
        )

    def submit_txn(self, client: jsonrpc.Client, script: diem_types.Script) -> diem_types.SignedTransaction:
        txn = self.create_txn(client, script)
        client.submit(txn)
        return txn

    def submit_and_wait_for_txn(self, client: jsonrpc.Client, script: diem_types.Script) -> jsonrpc.Transaction:
        txn = self.submit_txn(client, script)
        return client.wait_for_transaction(txn, timeout_secs=self.txn_expire_duration_secs)

    def rotate_dual_attestation_info(self, client: jsonrpc.Client, base_url: str) -> jsonrpc.Transaction:
        return self.submit_and_wait_for_txn(
            client,
            stdlib.encode_rotate_dual_attestation_info_script(
                new_url=base_url.encode("utf-8"), new_key=self.compliance_public_key_bytes
            ),
        )

    def to_dict(self) -> Dict[str, str]:
        """export to a string only dictionary for saving and importing as config

        private keys will be exported as hex-encded raw key bytes.
        """

        d = copy(self.__dict__)
        d["private_key"] = utils.private_key_bytes(self.private_key).hex()
        d["compliance_key"] = utils.private_key_bytes(self.compliance_key).hex()
        return d



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0


from diem.testing.miniwallet import RestClient, AccountResource
from diem import identifier, utils
from typing import Union
import pytest, requests


@pytest.fixture
def account(target_client: RestClient) -> AccountResource:
    return target_client.create_account()


def test_generate_account_payment_uri_without_currency_and_amount(account: AccountResource, hrp: str) -> None:
    uri = account.generate_payment_uri()
    assert uri.id
    assert uri.account_id == account.id
    assert uri.currency is None
    assert uri.amount is None
    assert uri.payment_uri

    intent = identifier.decode_intent(uri.payment_uri, hrp)
    assert intent.account_address
    assert intent.subaddress
    assert intent.currency_code is None
    assert intent.amount is None


def test_generate_account_payment_uri_with_currency(account: AccountResource, currency: str, hrp: str) -> None:
    uri = account.generate_payment_uri(currency=currency)
    assert uri.id
    assert uri.account_id == account.id
    assert uri.currency == currency
    assert uri.amount is None

    intent = identifier.decode_intent(uri.payment_uri, hrp)
    assert intent.account_address
    assert intent.subaddress
    assert intent.currency_code == currency
    assert intent.amount is None


def test_generate_account_payment_uri_with_amount(account: AccountResource, hrp: str) -> None:
    amount = 1123
    uri = account.generate_payment_uri(amount=amount)
    assert uri.id
    assert uri.account_id == account.id
    assert uri.currency is None
    assert uri.amount == amount

    intent = identifier.decode_intent(uri.payment_uri, hrp)
    assert intent.account_address
    assert intent.subaddress
    assert intent.currency_code is None
    assert intent.amount == amount


def test_generate_account_payment_uri_with_currency_and_amount(
    account: AccountResource, currency: str, hrp: str
) -> None:
    amount = 1123
    uri = account.generate_payment_uri(currency=currency, amount=amount)
    assert uri.id
    assert uri.account_id == account.id
    assert uri.currency == currency
    assert uri.amount == amount

    intent = identifier.decode_intent(uri.payment_uri, hrp)
    assert intent.account_address
    assert intent.subaddress
    assert intent.currency_code == currency
    assert intent.amount == amount


@pytest.mark.parametrize("invalid_currency", ["XU", "USD", "xus", "", '"XUS"', "X"])
def test_generate_account_payment_uri_with_invalid_currency(account: AccountResource, invalid_currency: str) -> None:
    with pytest.raises(requests.HTTPError, match="400 Client Error"):
        account.generate_payment_uri(currency=invalid_currency)


@pytest.mark.parametrize("amount", ["123", 12.34, -1, ""])
def test_generate_account_payment_uri_with_invalid_amount(account: AccountResource, amount: Union[str, int]) -> None:
    with pytest.raises(requests.HTTPError, match="400 Client Error"):
        account.generate_payment_uri(amount=amount)  # pyre-ignore


def test_generate_account_payment_uri_with_integer_overflow_amount(account: AccountResource) -> None:
    amount = 324234324342234234234234242234234324323423432432423423243242342342342342342342342334234
    with pytest.raises(requests.HTTPError, match="400 Client Error"):
        account.generate_payment_uri(amount=amount)


def test_generate_account_payment_URI_should_include_unique_subaddress(account: AccountResource, hrp: str) -> None:
    def subaddress(uri: str) -> str:
        return utils.hex(identifier.decode_intent(uri, hrp).subaddress)

    subaddresses = [subaddress(account.generate_payment_uri().payment_uri) for _ in range(10)]
    assert sorted(subaddresses) == sorted(list(set(subaddresses)))



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from .app import App
from .models import Account, Transaction, PaymentUri, Event, KycSample, Payment, PaymentCommand, RefundReason
from .falcon import falcon_api



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from diem import offchain
from diem.testing.miniwallet import RestClient
import pytest


def test_create_a_blank_account(target_client: RestClient) -> None:
    account = target_client.create_account()
    assert account.id
    assert account.kyc_data is None
    assert account.balance("XUS") == 0
    assert account.balance("XDX") == 0


@pytest.mark.parametrize("amount", [0, 1, 100, 10000000000])
@pytest.mark.parametrize("currency", ["XUS"])
def test_create_an_account_with_initial_deposit_balances(target_client: RestClient, currency: str, amount: int) -> None:
    account = target_client.create_account(kyc_data=None, balances={currency: amount})
    assert account.id
    assert account.kyc_data is None
    assert account.balance(currency) == amount


@pytest.mark.parametrize(  # pyre-ignore
    "kyc_data",
    [
        offchain.to_json(offchain.individual_kyc_data()),
        offchain.to_json(offchain.entity_kyc_data()),
    ],
)
def test_create_an_account_with_minimum_valid_kyc_data(target_client: RestClient, kyc_data: str) -> None:
    account = target_client.create_account(kyc_data=kyc_data)
    assert account.id
    assert account.kyc_data == kyc_data


def test_create_an_account_with_valid_kyc_data_and_initial_deposit_balances(
    target_client: RestClient, currency: str
) -> None:
    minimum_kyc_data = offchain.to_json(offchain.individual_kyc_data())
    account = target_client.create_account(kyc_data=minimum_kyc_data, balances={currency: 123})
    assert account.id
    assert account.kyc_data == minimum_kyc_data


def test_account_id_should_be_unique(target_client: RestClient) -> None:
    ids = [target_client.create_account().id for _ in range(10)]
    assert sorted(ids) == sorted(list(set(ids)))



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from diem import testnet
from diem.testing.miniwallet import AppConfig, ServerConfig
from diem.testing.suites import envs
from typing import Optional
import logging, click, functools, pytest, os, sys, re

log_format: str = "%(name)s [%(asctime)s] %(levelname)s: %(message)s"
click.option = functools.partial(click.option, show_default=True)  # pyre-ignore


@click.group()
def main() -> None:
    pass


@click.command()
@click.option("--name", "-n", default="mini-wallet", help="Application name.")
@click.option("--host", "-h", default="localhost", help="Start server host.")
@click.option("--port", "-p", default=8888, help="Start server port.")
@click.option("--jsonrpc", "-j", default=testnet.JSON_RPC_URL, help="Diem fullnode JSON-RPC URL.")
@click.option("--faucet", "-f", default=testnet.FAUCET_URL, help="Testnet faucet URL.")
@click.option("--logfile", "-l", default=None, help="Log to a file instead of printing into console.")
@click.option("--enable-debug-api", "-o", default=True, help="Enable debug API.", type=bool, is_flag=True)
def start_server(
    name: str, host: str, port: int, jsonrpc: str, faucet: str, enable_debug_api: bool, logfile: Optional[str] = None
) -> None:
    logging.basicConfig(level=logging.INFO, format=log_format, filename=logfile)
    configure_testnet(jsonrpc, faucet)

    conf = AppConfig(name=name, server_conf=ServerConfig(host=host, port=port), enable_debug_api=enable_debug_api)
    print("Server Config: %s" % conf)

    client = testnet.create_client()
    print("Diem chain id: %s" % client.get_metadata().chain_id)

    conf.start(client).join()


@click.command()
@click.option("--target", "-t", default="http://localhost:8888", help="Target mini-wallet application URL.")
@click.option("--stub-bind-host", "-h", default="localhost", help="The host the miniwallet stub server will bind to")
@click.option(
    "--stub-bind-port",
    "-p",
    default=None,
    help="The port the miniwallet stub server will bind to. Random if empty.",
    type=int,
)
@click.option(
    "--stub-diem-account-base-url",
    "-u",
    default=None,
    help="The address that will be used for offchain callbacks. Defaults to http://localhost:{random_port}",
)
@click.option("--jsonrpc", "-j", default=testnet.JSON_RPC_URL, help="Diem fullnode JSON-RPC URL.")
@click.option("--faucet", "-f", default=testnet.FAUCET_URL, help="Testnet faucet URL.")
@click.option(
    "--pytest-args",
    default="",
    help="Additional pytest arguments, split by empty space, e.g. `--pytest-args '-v -s'`.",
    show_default=False,
)
@click.option("--test-debug-api", "-d", default=False, help="Run tests for debug APIs.", type=bool)
@click.option("--verbose", "-v", default=False, help="Enable verbose log output.", type=bool, is_flag=True)
def test(
    target: str,
    stub_bind_host: Optional[str],
    stub_bind_port: Optional[int],
    stub_diem_account_base_url: Optional[str],
    jsonrpc: str,
    faucet: str,
    pytest_args: str,
    test_debug_api: bool,
    verbose: bool,
) -> None:
    configure_testnet(jsonrpc, faucet)
    os.environ[envs.TARGET_URL] = target

    # If stub_bind_host is provided, then stub_diem_account_base_url must be as well or the test won't work
    if stub_bind_host is not None and stub_bind_host != "localhost" and not stub_diem_account_base_url:
        raise click.ClickException("--stub-diem-account-base-url is required when passing in a custom --stub-bind-host")

    if stub_bind_host:
        os.environ[envs.DMW_STUB_BIND_HOST] = str(stub_bind_host)
    if stub_bind_port:
        os.environ[envs.DMW_STUB_BIND_PORT] = str(stub_bind_port)
    if stub_diem_account_base_url:
        os.environ[envs.DMW_STUB_DIEM_ACCOUNT_BASE_URL] = stub_diem_account_base_url

    if test_debug_api:
        os.environ[envs.TEST_DEBUG_API] = "Y"

    args = [arg for arg in re.compile("\\s+").split(pytest_args) if arg]
    args = ["--pyargs", "diem.testing.suites"] + args
    if verbose:
        args.append("--log-level=INFO")
        args.append("-s")
        args.append("-v")

    code = pytest.main(args)
    sys.stdout.flush()
    raise SystemExit(code)


def configure_testnet(jsonrpc: str, faucet: str) -> None:
    testnet.JSON_RPC_URL = jsonrpc
    testnet.FAUCET_URL = faucet
    print("Diem JSON-RPC URL: %s" % testnet.JSON_RPC_URL)
    print("Diem Testnet Faucet URL: %s" % testnet.FAUCET_URL)


main.add_command(start_server)
main.add_command(test)



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from diem.testing.miniwallet import Account, Transaction, RestClient
from .envs import should_test_debug_api, is_self_check
from .clients import Clients
import pytest, requests, time, json


def test_send_payment_and_events(clients: Clients, hrp: str, currency: str) -> None:
    receiver = clients.target.create_account()
    payment_uri = receiver.generate_payment_uri()

    amount = 1234
    sender = clients.stub.create_account(balances={currency: amount})
    assert sender.balance(currency) == amount

    index = len(sender.events())
    payment = sender.send_payment(currency, amount, payment_uri.intent(hrp).account_id)
    assert payment.account_id == sender.id
    assert payment.currency == currency
    assert payment.amount == amount
    assert payment.payee == payment_uri.intent(hrp).account_id

    sender.wait_for_balance(currency, 0)
    sender.wait_for_event("updated_transaction", status=Transaction.Status.completed, start_index=index)

    receiver.wait_for_balance(currency, amount)

    new_events = [e for e in sender.events(index) if e.type != "info"]
    assert len(new_events) == 5, new_events
    assert new_events[0].type == "created_transaction"
    assert new_events[1].type == "created_subaddress"
    assert new_events[2].type == "updated_transaction"
    assert sorted(list(json.loads(new_events[2].data).keys())) == ["id", "subaddress_hex"]
    assert new_events[3].type == "updated_transaction"
    assert sorted(list(json.loads(new_events[3].data).keys())) == ["id", "signed_transaction"]
    assert new_events[4].type == "updated_transaction"
    assert sorted(list(json.loads(new_events[4].data).keys())) == ["diem_transaction_version", "id", "status"]


def test_receive_payment_and_events(clients: Clients, currency: str, hrp: str) -> None:
    receiver = clients.stub.create_account()
    payment_uri = receiver.generate_payment_uri()

    index = len(receiver.events())
    amount = 1234
    sender = clients.target.create_account({currency: amount})
    payment = sender.send_payment(currency, amount, payment_uri.intent(hrp).account_id)

    receiver.wait_for_balance(currency, amount)
    sender.wait_for_balance(currency, 0)

    new_events = [e for e in receiver.events(index) if e.type != "info"]
    assert len(new_events) == 1
    assert new_events[0].type == "created_transaction"
    txn = Transaction(**json.loads(new_events[0].data))
    assert txn.id
    assert txn.currency == payment.currency
    assert txn.amount == payment.amount
    assert txn.diem_transaction_version


def test_receive_multiple_payments(clients: Clients, hrp: str, currency: str) -> None:
    receiver = clients.stub.create_account()
    payment_uri = receiver.generate_payment_uri()

    index = len(receiver.events())
    amount = 1234
    sender1 = clients.target.create_account({currency: amount})
    sender1.send_payment(currency, amount, payment_uri.intent(hrp).account_id)

    sender2 = clients.target.create_account({currency: amount})
    sender2.send_payment(currency, amount, payment_uri.intent(hrp).account_id)

    sender1.wait_for_balance(currency, 0)
    sender2.wait_for_balance(currency, 0)
    receiver.wait_for_balance(currency, amount * 2)

    new_events = [e for e in receiver.events(index) if e.type != "info"]
    assert len(new_events) == 2
    assert new_events[0].type == "created_transaction"
    assert new_events[1].type == "created_transaction"


@pytest.mark.parametrize(
    "invalid_payee",
    [
        "invalid id",
        "dm1p7ujcndcl7nudzwt8fglhx6wxnvqqqqqqqqqqqqqd8p9cq",
        "tdm1p7ujcndcl7nudzwt8fglhx6wxnvqqqqqqqqqqqqqv88j4x",
    ],
)
def test_send_payment_payee_is_invalid(clients: Clients, currency: str, invalid_payee: str, hrp: str) -> None:
    sender = clients.stub.create_account({currency: 100})

    index = len(sender.events())
    with pytest.raises(requests.exceptions.HTTPError, match="400 Client Error") as einfo:
        sender.send_payment(currency, 1, invalid_payee)

    if is_self_check():
        assert "'payee' is invalid account identifier" in einfo.value.response.text

    assert sender.balance(currency) == 100
    assert sender.events(index) == []


def test_return_client_error_if_send_payment_more_than_account_balance(
    clients: Clients, currency: str, hrp: str
) -> None:
    receiver = clients.target.create_account()
    payment_uri = receiver.generate_payment_uri()
    sender = clients.stub.create_account({currency: 100})

    index = len(sender.events())
    with pytest.raises(requests.exceptions.HTTPError, match="400 Client Error") as einfo:
        sender.send_payment(currency, 101, payment_uri.intent(hrp).account_id)
    if is_self_check():
        assert "account balance not enough" in einfo.value.response.text
    assert sender.balance(currency) == 100
    assert sender.events(index) == []


def test_send_payment_meets_travel_rule_limit(
    clients: Clients, currency: str, travel_rule_threshold: int, hrp: str
) -> None:
    amount = travel_rule_threshold
    receiver = clients.target.create_account()
    payment_uri = receiver.generate_payment_uri()
    sender = clients.stub.create_account({currency: amount}, kyc_data=clients.stub.new_kyc_data())
    payment = sender.send_payment(currency, amount, payee=payment_uri.intent(hrp).account_id)

    sender.wait_for_event("updated_transaction", id=payment.id, status=Transaction.Status.completed)
    sender.wait_for_balance(currency, 0)
    receiver.wait_for_balance(currency, travel_rule_threshold)


def test_account_balance_validation_should_exclude_canceled_transactions(
    clients: Clients, currency: str, travel_rule_threshold: int, hrp: str
) -> None:
    amount = travel_rule_threshold
    receiver = clients.target.create_account()
    payment_uri = receiver.generate_payment_uri()
    sender = clients.stub.create_account({currency: amount}, kyc_data=clients.target.new_reject_kyc_data())
    # payment should be rejected during offchain kyc data exchange
    payment = sender.send_payment(currency, amount, payee=payment_uri.intent(hrp).account_id)

    sender.wait_for_event("updated_transaction", id=payment.id, status=Transaction.Status.canceled)

    sender.send_payment(currency, travel_rule_threshold - 1, payment_uri.intent(hrp).account_id)

    receiver.wait_for_balance(currency, travel_rule_threshold - 1)
    sender.wait_for_balance(currency, 1)


@pytest.mark.parametrize(
    "amount",
    [
        0,
        1,
        1_000_000_000,
        1_000_000_000_000,
    ],
)
def test_internal_transfer(clients: Clients, currency: str, amount: int, hrp: str) -> None:
    receiver = clients.stub.create_account()
    payment_uri = receiver.generate_payment_uri()
    sender = clients.stub.create_account({currency: amount})

    index = len(sender.events())

    payment = sender.send_payment(currency, amount, payee=payment_uri.intent(hrp).account_id)
    assert payment.amount == amount
    assert payment.payee == payment_uri.intent(hrp).account_id

    sender.wait_for_event("updated_transaction", start_index=index, id=payment.id, status=Transaction.Status.completed)

    sender.wait_for_balance(currency, 0)
    receiver.wait_for_balance(currency, amount)


@pytest.mark.skipif(bool(not should_test_debug_api()), reason="test debug api is not enabled")
def test_create_account_event(target_client: RestClient, currency: str) -> None:
    before_timestamp = int(time.time() * 1000)
    account = target_client.create_account()
    after_timestamp = int(time.time() * 1000)

    events = account.events()
    assert len(events) == 1
    event = events[0]
    assert event.id
    assert event.timestamp >= before_timestamp
    assert event.timestamp <= after_timestamp
    assert event.type == "created_account"
    event_data = Account(**json.loads(event.data))
    assert event_data.kyc_data == account.kyc_data
    assert event_data.id == account.id


@pytest.mark.skipif(bool(not should_test_debug_api()), reason="test debug api is not enabled")
def test_create_account_payment_uri_events(target_client: RestClient, hrp: str) -> None:
    account = target_client.create_account()
    index = len(account.events())
    ret = account.generate_payment_uri()
    assert ret
    assert len(account.events(index)) == 1
    assert account.events(index)[0].type == "created_payment_uri"


@pytest.mark.skipif(bool(not is_self_check()), reason="self check is not enabled")
def test_openapi_spec(target_client: RestClient) -> None:
    resp = target_client.send("GET", "/openapi.yaml")
    assert resp.text



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from os import getenv, environ
from typing import Optional, Dict, Any


TARGET_URL: str = "DMW_TEST_TARGET"
TEST_DEBUG_API: str = "DMW_TEST_DEBUG_API"
SELF_CHECK: str = "DMW_SELF_CHECK"

DMW_STUB_BIND_HOST: str = "DMW_STUB_BIND_HOST"
DMW_STUB_BIND_PORT: str = "DMW_STUB_BIND_PORT"
DMW_STUB_DIEM_ACCOUNT_BASE_URL: str = "DMW_STUB_DIEM_ACCOUNT_BASE_URL"


def dmw_stub_diem_account_base_url() -> Optional[str]:
    return getenv(DMW_STUB_DIEM_ACCOUNT_BASE_URL)


def dmw_stub_server() -> Dict[str, Any]:
    return {
        k: v
        for k, v in {
            "host": dmw_stub_bind_host(),
            "port": dmw_stub_bind_port(),
            "base_url": dmw_stub_diem_account_base_url(),
        }.items()
        if v
    }


def dmw_stub_bind_host() -> Optional[str]:
    return getenv(DMW_STUB_BIND_HOST)


def dmw_stub_bind_port() -> Optional[int]:
    port = getenv(DMW_STUB_BIND_PORT)
    if port is not None:
        return int(port)


def target_url() -> str:
    return environ[TARGET_URL]


def is_self_check() -> bool:
    return getenv(SELF_CHECK) is not None


def should_test_debug_api() -> bool:
    return getenv(TEST_DEBUG_API) is not None



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0


from diem.testing.miniwallet import Event, AccountResource
from diem import offchain
from typing import List
from .clients import Clients
import pytest, json


@pytest.mark.parametrize("amount", [1, 999_999])
@pytest.mark.parametrize("actor", ["sender", "receiver"])
def test_payment_under_threshold_succeed(
    actor: str,
    amount: int,
    currency: str,
    clients: Clients,
    hrp: str,
) -> None:
    sender_client = clients.target if actor == "sender" else clients.stub
    receiver_client = clients.stub if actor == "sender" else clients.target

    sender = sender_client.create_account({currency: amount})
    receiver = receiver_client.create_account()
    sender_initial = sender.balance(currency)
    receiver_initial = receiver.balance(currency)

    payment_uri = receiver.generate_payment_uri()
    send_payment = sender.send_payment(currency, amount, payment_uri.intent(hrp).account_id)
    assert send_payment.currency == currency
    assert send_payment.amount == amount
    assert send_payment.payee == payment_uri.intent(hrp).account_id

    sender.wait_for_balance(currency, sender_initial - amount)
    receiver.wait_for_balance(currency, receiver_initial + amount)


@pytest.mark.parametrize(
    "sender_kyc, receiver_kyc, exchange_states, payment_result",
    [
        ("valid_kyc_data", "valid_kyc_data", ["S_INIT", "R_SEND", "READY"], "completed"),
        (
            "valid_kyc_data",
            "soft_match_kyc_data",
            ["S_INIT", "R_SEND", "S_SOFT", "R_SOFT_SEND", "READY"],
            "completed",
        ),
        ("valid_kyc_data", "reject_kyc_data", ["S_INIT", "R_SEND", "S_ABORT"], "failed"),
        (
            "valid_kyc_data",
            "soft_reject_kyc_data",
            ["S_INIT", "R_SEND", "S_SOFT", "R_SOFT_SEND", "S_ABORT"],
            "failed",
        ),
        (
            "soft_match_kyc_data",
            "valid_kyc_data",
            ["S_INIT", "R_SOFT", "S_SOFT_SEND", "R_SEND", "READY"],
            "completed",
        ),
        (
            "soft_match_kyc_data",
            "soft_match_kyc_data",
            ["S_INIT", "R_SOFT", "S_SOFT_SEND", "R_SEND", "S_SOFT", "R_SOFT_SEND", "READY"],
            "completed",
        ),
        (
            "soft_match_kyc_data",
            "reject_kyc_data",
            ["S_INIT", "R_SOFT", "S_SOFT_SEND", "R_SEND", "S_ABORT"],
            "failed",
        ),
        (
            "soft_match_kyc_data",
            "soft_reject_kyc_data",
            ["S_INIT", "R_SOFT", "S_SOFT_SEND", "R_SEND", "S_SOFT", "R_SOFT_SEND", "S_ABORT"],
            "failed",
        ),
        ("reject_kyc_data", "valid_kyc_data", ["S_INIT", "R_ABORT"], "failed"),
        ("reject_kyc_data", "soft_match_kyc_data", ["S_INIT", "R_ABORT"], "failed"),
        ("reject_kyc_data", "reject_kyc_data", ["S_INIT", "R_ABORT"], "failed"),
        ("reject_kyc_data", "soft_reject_kyc_data", ["S_INIT", "R_ABORT"], "failed"),
        (
            "soft_reject_kyc_data",
            "valid_kyc_data",
            ["S_INIT", "R_SOFT", "S_SOFT_SEND", "R_ABORT"],
            "failed",
        ),
        (
            "soft_reject_kyc_data",
            "soft_match_kyc_data",
            ["S_INIT", "R_SOFT", "S_SOFT_SEND", "R_ABORT"],
            "failed",
        ),
        (
            "soft_reject_kyc_data",
            "reject_kyc_data",
            ["S_INIT", "R_SOFT", "S_SOFT_SEND", "R_ABORT"],
            "failed",
        ),
        (
            "soft_reject_kyc_data",
            "soft_reject_kyc_data",
            ["S_INIT", "R_SOFT", "S_SOFT_SEND", "R_ABORT"],
            "failed",
        ),
    ],
)
@pytest.mark.parametrize("actor", ["sender", "receiver"])
def test_payment_meets_travel_rule_threshold(
    actor: str,
    sender_kyc: str,
    receiver_kyc: str,
    exchange_states: List[str],
    payment_result: str,
    currency: str,
    travel_rule_threshold: int,
    clients: Clients,
    hrp: str,
) -> None:
    amount = travel_rule_threshold
    sender_client = clients.target if actor == "sender" else clients.stub
    receiver_client = clients.stub if actor == "sender" else clients.target

    sender = sender_client.create_account(
        {currency: amount}, kyc_data=getattr(receiver_client, "new_%s" % sender_kyc)()
    )
    receiver = receiver_client.create_account(kyc_data=getattr(sender_client, "new_%s" % receiver_kyc)())
    sender_initial = sender.balance(currency)
    receiver_initial = receiver.balance(currency)

    payment_uri = receiver.generate_payment_uri()
    send_payment = sender.send_payment(currency, amount, payment_uri.intent(hrp).account_id)
    assert send_payment.currency == currency
    assert send_payment.amount == amount
    assert send_payment.payee == payment_uri.intent(hrp).account_id

    stub_account: AccountResource = sender if clients.stub == sender.client else receiver

    def match_exchange_states() -> None:
        assert payment_command_event_states(stub_account) == exchange_states

    stub_account.wait_for(match_exchange_states)

    if payment_result == "completed":
        sender.wait_for_balance(currency, sender_initial - amount)
        receiver.wait_for_balance(currency, receiver_initial + amount)
    else:
        sender.wait_for_balance(currency, sender_initial)
        receiver.wait_for_balance(currency, receiver_initial)


def payment_state_id(event: Event) -> str:
    payment = offchain.from_dict(json.loads(event.data)["payment_object"], offchain.PaymentObject)
    return offchain.payment_state.MACHINE.match_state(payment).id


def payment_command_event_states(account: AccountResource) -> List[str]:
    return [payment_state_id(event) for event in account.events() if is_payment_command_event(event)]


def is_payment_command_event(e: Event) -> bool:
    return e.type in ["created_payment_command", "updated_payment_command"]



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from copy import copy
from dataclasses import dataclass, field
from typing import Dict, Callable, Any
from .store import InMemoryStore, NotFoundError
from .models import Transaction, Subaddress, PaymentCommand, RefundReason
from .... import jsonrpc, diem_types, txnmetadata, identifier, utils


PENDING_INBOUND_ACCOUNT_ID: str = "pending_inbound_account"


@dataclass
class EventPuller:
    client: jsonrpc.Client
    store: InMemoryStore
    hrp: str
    state: Dict[str, int] = field(default_factory=dict)

    def add(self, address: diem_types.AccountAddress) -> None:
        account = self.client.must_get_account(address)
        self.state[account.received_events_key] = 0

    def fetch(self, process: Callable[[jsonrpc.Event], None], batch_size: int = 100) -> None:
        for key, seq in self.state.items():
            events = self.client.get_events(key, seq, batch_size)
            if events:
                for event in events:
                    process(event)
                    self.state[key] = event.sequence_number + 1

    def head(self) -> None:
        state = None
        while state != self.state:
            state = copy(self.state)
            self.fetch(lambda _: None)

    def save_payment_txn(self, event: jsonrpc.Event) -> None:
        try:
            self._save_payment_txn(event)
        except (NotFoundError, ValueError) as e:
            self._create_txn(PENDING_INBOUND_ACCOUNT_ID, event)
            self.store.create_event(PENDING_INBOUND_ACCOUNT_ID, "info", str(e))

    def _save_payment_txn(self, event: jsonrpc.Event) -> None:
        metadata = txnmetadata.decode_structure(event.data.metadata)
        if isinstance(metadata, diem_types.GeneralMetadataV0):
            try:
                res = self.store.find(Subaddress, subaddress_hex=utils.hex(metadata.to_subaddress))
                return self._create_txn(res.account_id, event, subaddress_hex=res.subaddress_hex)
            except NotFoundError:
                self._create_txn(PENDING_INBOUND_ACCOUNT_ID, event)
                payee = identifier.encode_account(event.data.sender, metadata.from_subaddress, self.hrp)
                return self._refund(payee, RefundReason.invalid_subaddress, event)
        elif isinstance(metadata, diem_types.TravelRuleMetadataV0) and metadata.off_chain_reference_id:
            cmd = self.store.find(PaymentCommand, reference_id=metadata.off_chain_reference_id)
            return self._create_txn(cmd.account_id, event, reference_id=cmd.reference_id)
        elif isinstance(metadata, diem_types.RefundMetadataV0):
            version = int(metadata.transaction_version)
            reason = RefundReason.from_diem_type(metadata.reason)
            diem_txn = self.client.get_transactions(version, 1)[0]
            if diem_txn.transaction.script and diem_txn.transaction.script.metadata:
                original_metadata = txnmetadata.decode_structure(diem_txn.transaction.script.metadata)
                if isinstance(original_metadata, diem_types.GeneralMetadataV0):
                    original_sender_subaddress = utils.hex(original_metadata.from_subaddress)
                    sub = self.store.find(Subaddress, subaddress_hex=original_sender_subaddress)
                    return self._create_txn(
                        sub.account_id, event, refund_diem_txn_version=version, refund_reason=reason
                    )
        raise ValueError("unrecognized metadata: %r" % event.data.metadata)

    def _refund(self, payee: str, reason: RefundReason, event: jsonrpc.Event) -> None:
        self.store.create(
            Transaction,
            account_id=PENDING_INBOUND_ACCOUNT_ID,
            status=Transaction.Status.pending,
            currency=event.data.amount.currency,
            amount=event.data.amount.amount,
            payee=payee,
            refund_diem_txn_version=event.transaction_version,
            refund_reason=reason,
        )

    def _create_txn(self, account_id: str, event: jsonrpc.Event, **kwargs: Any) -> None:
        self.store.create(
            Transaction,
            account_id=account_id,
            currency=event.data.amount.currency,
            amount=event.data.amount.amount,
            diem_transaction_version=event.transaction_version,
            status=Transaction.Status.completed,
            **kwargs,
        )



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass, field, asdict
from typing import Dict, Any
from .client import RestClient
from .app import App, falcon_api
from ... import offchain, testnet, jsonrpc, LocalAccount
import waitress, threading, logging, falcon, json


@dataclass
class ServerConfig:
    host: str = field(default="localhost")
    port: int = field(default_factory=offchain.http_server.get_available_port)
    base_url: str = field(default="")

    def __post_init__(self) -> None:
        if not self.base_url:
            self.base_url = f"http://localhost:{self.port}"


@dataclass
class AppConfig:
    name: str = field(default="mini-wallet")
    enable_debug_api: bool = field(default=False)
    account_config: Dict[str, Any] = field(default_factory=lambda: LocalAccount().to_dict())
    server_conf: ServerConfig = field(default_factory=ServerConfig)
    initial_amount: int = field(default=1_000_000_000_000)
    initial_currency: str = field(default=testnet.TEST_CURRENCY_CODE)

    @property
    def logger(self) -> logging.Logger:
        return logging.getLogger(self.name)

    @property
    def account(self) -> LocalAccount:
        return LocalAccount.from_dict(self.account_config)

    @property
    def server_url(self) -> str:
        return self.server_conf.base_url

    def create_client(self) -> RestClient:
        self.logger.info(f"Creating client pointing to {self.server_url}")
        return RestClient(server_url=self.server_url, name="%s-client" % self.name).with_retry()

    def setup_account(self, client: jsonrpc.Client) -> None:
        acc = client.get_account(self.account.account_address)
        if not acc or self.need_funds(acc):
            self.logger.info("faucet mint %s" % self.account.account_address.to_hex())
            faucet = testnet.Faucet(client)
            faucet.mint(self.account.auth_key.hex(), self.initial_amount, self.initial_currency)
        if not acc or self.need_rotate(acc):
            self.logger.info("rotate dual attestation info for %s" % self.account.account_address.to_hex())
            self.logger.info("set base url to: %s" % self.server_url)
            self.account.rotate_dual_attestation_info(client, self.server_url)

    def need_funds(self, account: jsonrpc.Account) -> bool:
        for balance in account.balances:
            if balance.currency == self.initial_currency and balance.amount > self.initial_amount / 2:
                return False
        return True

    def need_rotate(self, account: jsonrpc.Account) -> bool:
        if account.role.base_url != self.server_url:
            return True
        if not account.role.compliance_key:
            return True
        if bytes.fromhex(account.role.compliance_key) != self.account.compliance_public_key_bytes:
            return True
        return False

    def serve(self, client: jsonrpc.Client) -> threading.Thread:
        api: falcon.API = falcon_api(App(self.account, client, self.name, self.logger), self.enable_debug_api)

        def serve() -> None:
            self.logger.info(f"serving on {self.server_conf.host}:{self.server_conf.port} at {self.server_url}")
            waitress.serve(
                api,
                host=self.server_conf.host,
                port=self.server_conf.port,
                clear_untrusted_proxy_headers=True,
                _quiet=True,
            )

        t = threading.Thread(target=serve, daemon=True)
        t.start()
        return t

    def start(self, client: jsonrpc.Client) -> threading.Thread:
        self.setup_account(client)
        return self.serve(client)

    def __str__(self) -> str:
        return json.dumps(asdict(self), indent=2)



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass, field, asdict
from typing import List, Dict, Type, Any, Callable, TypeVar, Generator
from .models import Base, Account, Event
from .... import utils
import json, time, threading


T = TypeVar("T", bound=Base)


class NotFoundError(ValueError):
    pass


@dataclass
class InMemoryStore:
    """InMemoryStore is a simple in-memory store for resources"""

    resources: Dict[Type[Base], List[Dict[str, Any]]] = field(default_factory=dict)
    resources_lock: threading.Lock = field(default_factory=threading.Lock)
    gen_id_lock: threading.Lock = field(default_factory=threading.Lock)
    gen_id: int = field(default=0)

    def next_id(self) -> int:
        with self.gen_id_lock:
            self.gen_id += 1
            return self.gen_id

    def create_event(self, account_id: str, type: str, data: str) -> Event:
        return self.create(Event, account_id=account_id, type=type, data=data, timestamp=_ts())

    def find(self, typ: Type[T], **conds: Any) -> T:
        list = self._select(typ, **conds)
        ret = next(list, None)
        if not ret:
            raise NotFoundError("%s not found by %s" % (typ.__name__, conds))
        if next(list, None):
            raise ValueError("found multiple resources data matches %s" % conds)
        return ret

    def find_all(self, typ: Type[T], **conds: Any) -> List[T]:
        return list(self._select(typ, **conds))

    def create(self, typ: Type[T], before_create: Callable[[Dict[str, Any]], None] = lambda _: _, **data: Any) -> T:
        with self.resources_lock:
            before_create(data)
            obj = typ(**self._insert(typ, **data))
            self._record_event(obj, "created", data)
            return obj

    def update(self, obj: T, before_update: Callable[[T], None] = lambda _: _, **data: Any) -> None:
        for k, v in data.items():
            setattr(obj, k, v)
        with self.resources_lock:
            before_update(obj)
            self._update(obj)
            self._record_event(obj, "updated", data)

    def _record_event(self, obj: T, action: str, data: Dict[str, Any]) -> None:
        if not isinstance(obj, Event):
            type = "%s_%s" % (action, utils.to_snake(obj))
            account_id = obj.id if isinstance(obj, Account) else obj.account_id  # pyre-ignore
            data["id"] = obj.id
            self._insert(Event, account_id=account_id, type=type, data=json.dumps(data), timestamp=_ts())

    def _update(self, obj: T) -> None:
        records = self.resources.get(type(obj), [])
        index = next(iter([i for i, res in enumerate(records) if res["id"] == obj.id]), None)
        if index is None:
            raise NotFoundError("could not find resource by id: %s" % obj.id)
        records[index] = asdict(obj)

    def _insert(self, typ: Type[T], **res: Any) -> Dict[str, Any]:
        res["id"] = str(self.next_id())
        self.resources.setdefault(typ, []).append(asdict(typ(**res)))
        return res

    def _select(self, typ: Type[T], reverse: bool = False, **conds: Any) -> Generator[T, None, None]:
        items = reversed(self.resources.get(typ, [])) if reverse else self.resources.get(typ, [])
        for res in items:
            if _match(res, **conds):
                yield typ(**res)


def _match(res: Dict[str, Any], **conds: Any) -> bool:
    for k, v in conds.items():
        if res.get(k) != v:
            return False
    return True


def _ts() -> int:
    return int(time.time() * 1000)



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
from typing import Optional, Type, TypeVar, Dict, Callable, Any


T = TypeVar("T")
Validator = Callable[[str, T], None]


@dataclass
class JsonInput:
    data: Dict[str, Any]

    def get_nullable(self, name: str, typ: Type[T], validator: Optional[Validator] = None) -> Optional[T]:
        val = self.data.get(name, None)
        if val is None:
            return None
        if isinstance(val, typ):
            if validator:
                validator(name, val)
            return val
        raise ValueError("%r type must be %r, but got %r" % (name, typ.__name__, type(val).__name__))

    def get(self, name: str, typ: Type[T], validator: Optional[Validator] = None) -> T:
        val = self.get_nullable(name, typ, validator)
        if val is None:
            raise ValueError("%r is required" % name)
        return val



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0


from ... import testnet, jsonrpc
from ..miniwallet import RestClient, AppConfig, AccountResource, ServerConfig
from ..miniwallet.app.event_puller import PENDING_INBOUND_ACCOUNT_ID
from .clients import Clients
from .envs import (
    target_url,
    is_self_check,
    should_test_debug_api,
    dmw_stub_server,
)
import pytest


@pytest.fixture(scope="package")
def target_client(diem_client: jsonrpc.Client) -> RestClient:
    if is_self_check():
        conf = AppConfig(name="target-wallet", enable_debug_api=should_test_debug_api())
        print("self-checking, launch target app with config %s" % conf)
        conf.start(diem_client)
        return conf.create_client()
    print("target wallet server url: %s" % target_url())
    return RestClient(name="target-wallet-client", server_url=target_url()).with_retry()


@pytest.fixture(scope="package")
def diem_client() -> jsonrpc.Client:
    print("Diem JSON-RPC URL: %s" % testnet.JSON_RPC_URL)
    print("Diem Testnet Faucet URL: %s" % testnet.FAUCET_URL)
    return testnet.create_client()


@pytest.fixture(scope="package")
def stub_config() -> AppConfig:
    return AppConfig(name="stub-wallet", enable_debug_api=True, server_conf=ServerConfig(**dmw_stub_server()))


@pytest.fixture(scope="package")
def stub_client(stub_config: AppConfig, diem_client: jsonrpc.Client) -> RestClient:
    print("Start stub app with config %s" % stub_config)
    stub_config.start(diem_client)
    return stub_config.create_client()


@pytest.fixture(scope="package")
def clients(stub_client: RestClient, target_client: RestClient, diem_client: jsonrpc.Client) -> Clients:
    return Clients(
        target=target_client,
        stub=stub_client,
        diem=diem_client,
    )


@pytest.fixture
def hrp() -> str:
    return testnet.HRP


@pytest.fixture
def currency() -> str:
    return testnet.TEST_CURRENCY_CODE


@pytest.fixture
def travel_rule_threshold(clients: Clients) -> int:
    # todo: convert the limit base on currency
    return clients.diem.get_metadata().dual_attestation_limit


@pytest.fixture
def pending_income_account(stub_client: RestClient) -> AccountResource:
    return AccountResource(id=PENDING_INBOUND_ACCOUNT_ID, client=stub_client)



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from diem.testing.miniwallet import RestClient, AccountResource, Transaction, AppConfig, RefundReason
from diem import jsonrpc, stdlib, utils, txnmetadata, diem_types
from typing import Generator, Optional
import pytest


amount: int = 123


@pytest.fixture
def sender_account(
    stub_client: RestClient, currency: str, pending_income_account: AccountResource
) -> Generator[AccountResource, None, None]:
    account = stub_client.create_account(balances={currency: amount})
    yield account
    account.log_events()
    # MiniWallet stub saves the payment without account information (subaddress / reference id)
    # into a pending income account before process it.
    # Here we log events of the account for showing more context related to the test
    # when test failed.
    pending_income_account.log_events()


@pytest.fixture
def receiver_account(target_client: RestClient) -> Generator[AccountResource, None, None]:
    account = target_client.create_account()
    yield account
    account.log_events()


@pytest.mark.parametrize("invalid_metadata", [b"", b"invalid metadata"])
def test_receive_payment_with_invalid_metadata(
    sender_account: AccountResource,
    receiver_account: AccountResource,
    currency: str,
    hrp: str,
    stub_config: AppConfig,
    diem_client: jsonrpc.Client,
    invalid_metadata: bytes,
) -> None:
    """When received a payment with invalid metadata, it is up to the wallet application how to handle it.
    This test makes sure target wallet application should continue to process valid transactions after
    received such an on-chain transaction.

    Test Plan:
    1. Generate a valid payment URI from receiver account.
    2. Submit a p2p transaction with invalid metadata, and wait for it is executed.
    3. Send a valid payment to the payment URI.
    4. Assert receiver account received the valid payment.
    """

    uri = receiver_account.generate_payment_uri()
    receiver_account_address = uri.intent(hrp).account_address
    stub_config.account.submit_and_wait_for_txn(
        diem_client,
        stdlib.encode_peer_to_peer_with_metadata_script(
            currency=utils.currency_code(currency),
            amount=amount * 2,
            payee=receiver_account_address,
            metadata=invalid_metadata,
            metadata_signature=b"",
        ),
    )
    assert receiver_account.balance(currency) == 0

    pay = sender_account.send_payment(currency=currency, amount=amount, payee=uri.intent(hrp).account_id)
    wait_for_payment_transaction_complete(sender_account, pay.id)
    receiver_account.wait_for_balance(currency, amount)


def test_receive_payment_with_general_metadata_and_valid_from_and_to_subaddresses(
    sender_account: AccountResource, receiver_account: AccountResource, currency: str, hrp: str
) -> None:
    """
    Test Plan:
    1. Generate a valid payment URI from receiver account.
    2. Send a payment to the payee from the valid payment URI.
    4. Wait for the transaction executed successfully.
    5. Assert receiver account received the fund.
    """
    uri = receiver_account.generate_payment_uri()
    pay = sender_account.send_payment(currency=currency, amount=amount, payee=uri.intent(hrp).account_id)
    wait_for_payment_transaction_complete(sender_account, pay.id)
    receiver_account.wait_for_balance(currency, amount)


@pytest.mark.parametrize(  # pyre-ignore
    "invalid_to_subaddress", [None, b"", b"bb4a3ba109a3175f", b"subaddress_more_than_8_bytes", b"too_short"]
)
def test_receive_payment_with_general_metadata_and_invalid_to_subaddress(
    sender_account: AccountResource,
    receiver_account: AccountResource,
    currency: str,
    hrp: str,
    stub_config: AppConfig,
    diem_client: jsonrpc.Client,
    invalid_to_subaddress: Optional[bytes],
) -> None:
    """When received a payment with general metadata and invalid to subaddress,
    receiver should refund the payment by using RefundMetadata with reason `invalid subaddress`.

    Test Plan:
    1. Generate a valid payment URI from the receiver account.
    2. Create a general metadata with valid from subaddress and invalid to subaddress.
    3. Send payment transaction from sender to receiver on-chain account.
    4. Wait for the transaction executed successfully.
    5. Assert sender account received a payment transaction with refund metadata.
    6. Assert receiver account does not receive funds.
    """

    receiver_uri = receiver_account.generate_payment_uri()
    receiver_account_address: diem_types.AccountAddress = receiver_uri.intent(hrp).account_address

    sender_uri = sender_account.generate_payment_uri()
    valid_from_subaddress = sender_uri.intent(hrp).subaddress
    invalid_metadata = txnmetadata.general_metadata(valid_from_subaddress, invalid_to_subaddress)
    original_payment_txn: jsonrpc.Transaction = stub_config.account.submit_and_wait_for_txn(
        diem_client,
        stdlib.encode_peer_to_peer_with_metadata_script(
            currency=utils.currency_code(currency),
            amount=amount,
            payee=receiver_account_address,
            metadata=invalid_metadata,
            metadata_signature=b"",
        ),
    )

    sender_account.wait_for_event(
        "created_transaction",
        status=Transaction.Status.completed,
        refund_diem_txn_version=original_payment_txn.version,
        refund_reason=RefundReason.invalid_subaddress,
    )
    assert receiver_account.balance(currency) == 0


def wait_for_payment_transaction_complete(account: AccountResource, payment_id: str) -> None:
    # MiniWallet stub generates `updated_transaction` event when transaction is completed on-chain
    # Payment id is same with Transaction id.
    account.wait_for_event("updated_transaction", status=Transaction.Status.completed, id=payment_id)



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from .app import (
    App,
    Account,
    Transaction,
    PaymentUri,
    Event,
    Payment,
    PaymentCommand,
    KycSample,
    RefundReason,
    falcon_api,
)
from .client import RestClient, AccountResource
from .config import AppConfig, ServerConfig



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass, field, replace, asdict
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from typing import List, Optional, Any, Dict, Callable
from .app import PaymentUri, KycSample, Event, Payment
from .app.store import _match
from ... import offchain, jsonrpc
import requests, logging, random, string, json, time


@dataclass
class RestClient:
    name: str
    server_url: str
    session: requests.Session = field(default_factory=requests.Session)
    logger: logging.Logger = field(init=False)

    def __post_init__(self) -> None:
        self.logger = logging.getLogger(self.name)

    def with_retry(self, retry: Retry = Retry(total=5, connect=5, backoff_factor=0.01)) -> "RestClient":
        self.session.mount(self.server_url, HTTPAdapter(max_retries=retry))
        return self

    def create_account(
        self, balances: Optional[Dict[str, int]] = None, kyc_data: Optional[str] = None
    ) -> "AccountResource":
        account = self.create("/accounts", kyc_data=kyc_data, balances=balances)
        return AccountResource(client=self, id=account["id"], kyc_data=account.get("kyc_data", None))

    def new_soft_match_kyc_data(self) -> str:
        return self.new_kyc_data(sample="soft_match")

    def new_reject_kyc_data(self) -> str:
        return self.new_kyc_data(sample="reject")

    def new_soft_reject_kyc_data(self) -> str:
        return self.new_kyc_data(sample="soft_reject")

    def new_valid_kyc_data(self) -> str:
        return self.new_kyc_data(sample="minimum")

    def new_kyc_data(self, name: Optional[str] = None, sample: str = "minimum") -> str:
        obj = offchain.from_json(getattr(self.kyc_sample(), sample), offchain.KycDataObject)
        if not name:
            name = "".join(random.choices(string.ascii_uppercase + string.digits, k=6))
        return offchain.to_json(replace(obj, legal_entity_name=name))

    def kyc_sample(self) -> KycSample:
        return KycSample(**self.send("GET", "/kyc_sample").json())

    def create(self, path: str, **kwargs: Any) -> Dict[str, Any]:
        return self.send("POST", path, json.dumps(kwargs) if kwargs else None).json()

    def get(self, path: str) -> Dict[str, Any]:
        return self.send("GET", path).json()

    def send(self, method: str, path: str, data: Optional[str] = None) -> requests.Response:
        url = "%s/%s" % (self.server_url.rstrip("/"), path.lstrip("/"))
        self.logger.info("%s %s: %s" % (method, path, data))
        resp = self.session.request(
            method=method,
            url=url.lower(),
            data=data,
            headers={"Content-Type": "application/json", "User-Agent": jsonrpc.client.USER_AGENT_HTTP_HEADER},
        )
        self.logger.info("response status code: %s" % resp.status_code)
        self.logger.info("response body: %s" % resp.text)
        resp.raise_for_status()
        return resp


@dataclass
class AccountResource:

    client: RestClient
    id: str
    kyc_data: Optional[str] = field(default=None)

    def balance(self, currency: str) -> int:
        return self._balances().get(currency, 0)

    def send_payment(self, currency: str, amount: int, payee: str) -> Payment:
        p = self.client.create(self._resources("payment"), payee=payee, currency=currency, amount=amount)
        return Payment(**p)

    def generate_payment_uri(self, currency: Optional[str] = None, amount: Optional[int] = None) -> PaymentUri:
        return PaymentUri(**self.client.create(self._resources("payment_uri"), currency=currency, amount=amount))

    def events(self, start: int = 0) -> List[Event]:
        ret = self.client.send("GET", self._resources("event")).json()
        return [Event(**obj) for obj in ret[start:]]

    def wait_for_balance(self, currency: str, amount: int) -> None:
        def fn() -> None:
            assert self.balance(currency) == amount, "currency: %s" % currency

        self.wait_for(fn)

    def wait_for_event(self, event_type: str, start_index: int = 0, **kwargs: Any) -> None:
        def fn() -> None:
            events = [e for e in self.events(start_index) if e.type == event_type]
            for e in events:
                if _match(json.loads(e.data), **kwargs):
                    return
            assert False, "could not find %s event with %s" % (event_type, kwargs)

        self.wait_for(fn)

    def wait_for(self, fn: Callable[[], None], max_tries: int = 120, delay: float = 0.1) -> None:
        tries = 0
        while True:
            tries += 1
            try:
                return fn()
            except AssertionError as e:
                if tries >= max_tries:
                    raise TimeoutError("%s, events: \n%s" % (e, self.dump_events())) from e
                time.sleep(delay)

    def log_events(self) -> None:
        events = self.dump_events()
        if events:
            self.client.logger.info("account(%s) events: %s", self.id, events)

    def dump_events(self) -> str:
        try:
            return json.dumps(list(map(self.event_asdict, self.events())), indent=2)
        except requests.HTTPError:
            return ""

    def event_asdict(self, event: Event) -> Dict[str, Any]:
        ret = asdict(event)
        try:
            ret["data"] = json.loads(event.data)
            if ret["data"].get("kyc_data"):
                ret["data"]["kyc_data"] = json.loads(ret["data"]["kyc_data"])
        except json.decoder.JSONDecodeError:
            pass
        return ret

    def info(self, *args: Any, **kwargs: Any) -> None:
        self.client.logger.info(*args, **kwargs)

    def _balances(self) -> Dict[str, int]:
        """returns account balances object

        should always prefer to use func `balance(currency) -> int`, which returns zero
        when currency not exist in the response.
        """

        return self.client.get(self._resources("balance"))

    def _resources(self, resource: str) -> str:
        return "/accounts/%s/%ss" % (self.id, resource)



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
from diem import jsonrpc
from ..miniwallet import RestClient


@dataclass
class Clients:
    target: RestClient
    stub: RestClient
    diem: jsonrpc.Client



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass, asdict
from typing import Dict, List, Any, Tuple
from .app import App
from .json_input import JsonInput
from .... import offchain
import falcon, json, traceback, logging, pathlib


base_path: str = str(pathlib.Path(__file__).resolve().parent.joinpath("static"))


@dataclass
class LoggerMiddleware:
    logger: logging.Logger

    def process_request(self, req, resp):  # pyre-ignore
        self.logger.info("%s %s" % (req.method, req.relative_uri))

    def process_response(self, req, resp, *args, **kwargs):  # pyre-ignore
        self.logger.info(resp.status)


def rest_handler(fn: Any):  # pyre-ignore
    def wrapper(self, req, resp, **kwargs):  # pyre-ignore
        try:
            try:
                data = json.load(req.stream)
                self.app.logger.info("request body: %s" % data)
            except Exception:
                data = {}
            status, body = fn(self, input=JsonInput(data), **kwargs)
            resp.status = status
            resp.body = json.dumps(body)
        except ValueError as e:
            resp.status = falcon.HTTP_400
            resp.body = json.dumps({"error": str(e), "stacktrace": traceback.format_exc()})

    return wrapper


@dataclass
class Endpoints:
    app: App

    @rest_handler
    def on_post_accounts(self, input: JsonInput) -> Tuple[str, Dict[str, str]]:
        return (falcon.HTTP_201, asdict(self.app.create_account(input)))

    @rest_handler
    def on_post_payments(self, account_id: str, input: JsonInput) -> Tuple[str, Dict[str, Any]]:
        return (falcon.HTTP_202, asdict(self.app.create_account_payment(account_id, input)))

    @rest_handler
    def on_post_payment_uris(self, account_id: str, input: JsonInput) -> Tuple[str, Dict[str, Any]]:
        return (falcon.HTTP_200, asdict(self.app.create_account_payment_uri(account_id, input)))

    @rest_handler
    def on_get_balances(self, account_id: str, input: JsonInput) -> Tuple[str, Dict[str, int]]:
        return (falcon.HTTP_200, self.app.get_account_balances(account_id))

    @rest_handler
    def on_get_events(self, account_id: str, input: JsonInput) -> Tuple[str, List[Dict[str, Any]]]:
        return (falcon.HTTP_200, [asdict(e) for e in self.app.get_account_events(account_id)])

    @rest_handler
    def on_get_kyc_sample(self, input: JsonInput) -> Tuple[str, Dict[str, str]]:
        return (falcon.HTTP_200, asdict(self.app.kyc_sample))

    def on_post_offchain(self, req: falcon.Request, resp: falcon.Response) -> None:
        request_id = req.get_header(offchain.X_REQUEST_ID)
        resp.set_header(offchain.X_REQUEST_ID, request_id)
        request_sender_address = req.get_header(offchain.X_REQUEST_SENDER_ADDRESS)
        input_data = req.stream.read()
        try:
            resp_obj = self.app.offchain_api(request_sender_address, input_data)
        except offchain.Error as e:
            self.app.logger.info(input_data)
            self.app.logger.exception(e)
            resp_obj = offchain.reply_request(cid=None, err=e.obj)
            resp.status = falcon.HTTP_400
        resp.body = self.app.jws_serialize(resp_obj)

    def on_get_index(self, req: falcon.Request, resp: falcon.Response) -> None:
        raise falcon.HTTPMovedPermanently("/index.html")


def falcon_api(app: App, enable_debug_api: bool = False) -> falcon.API:
    endpoints = Endpoints(app=app)
    api = falcon.API(middleware=[LoggerMiddleware(logger=app.logger)])
    api.add_static_route("/", base_path)
    api.add_route("/", endpoints, suffix="index")
    api.add_route("/accounts", endpoints, suffix="accounts")
    for res in ["balances", "payments", "payment_uris", "events"]:
        if res != "events" or enable_debug_api:
            api.add_route("/accounts/{account_id}/%s" % res, endpoints, suffix=res)
    api.add_route("/kyc_sample", endpoints, suffix="kyc_sample")
    api.add_route("/v2/command", endpoints, suffix="offchain")
    app.start_sync()
    return api



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import asdict
from typing import List, Tuple, Dict, Optional, Any
from json.decoder import JSONDecodeError
from .store import InMemoryStore, NotFoundError
from .diem_account import DiemAccount
from .models import PaymentUri, Subaddress, Account, Transaction, Event, KycSample, Payment, PaymentCommand
from .event_puller import EventPuller
from .json_input import JsonInput
from .... import jsonrpc, offchain, utils, LocalAccount, identifier
from ....offchain import KycDataObject, Status, AbortCode, CommandResponseObject
import threading, logging, numpy


class Base:
    def __init__(self, account: LocalAccount, client: jsonrpc.Client, name: str, logger: logging.Logger) -> None:
        self.logger = logger
        self.diem_account = DiemAccount(account, client)
        self.store = InMemoryStore()
        self.offchain = offchain.Client(account.account_address, client, account.hrp)
        self.kyc_sample: KycSample = KycSample.gen(name)
        self.event_puller = EventPuller(client=client, store=self.store, hrp=account.hrp)
        self.event_puller.add(account.account_address)
        self.event_puller.head()

    def _validate_kyc_data(self, name: str, val: str) -> None:
        try:
            offchain.from_json(val, KycDataObject)
        except (JSONDecodeError, offchain.types.FieldError) as e:
            raise ValueError("%r must be JSON-encoded KycDataObject: %s" % (name, e))

    def _validate_currency_code(self, name: str, val: str) -> None:
        try:
            self.offchain.validate_currency_code(val)
        except ValueError:
            raise ValueError("%r is invalid currency code: %s" % (name, val))

    def _validate_account_identifier(self, name: str, val: str) -> None:
        try:
            self.diem_account.account.decode_account_identifier(val)
        except ValueError as e:
            raise ValueError("%r is invalid account identifier: %s" % (name, e))

    def _validate_amount(self, name: str, val: int) -> None:
        if val < 0:
            raise ValueError("%r value must be greater than or equal to zero" % name)
        try:
            numpy.uint64(val)
        except OverflowError:
            raise ValueError("%r value is too big" % name)

    def _validate_account_balance(self, txn: Dict[str, Any]) -> None:
        if txn.get("payee"):
            balance = self._balances(txn["account_id"]).get(txn["currency"], 0)
            if balance < txn["amount"]:
                raise ValueError("account balance not enough: %s (%s)" % (balance, txn))

    def _balances(self, account_id: str) -> Dict[str, int]:
        ret = {}
        for txn in self.store.find_all(Transaction, account_id=account_id):
            if txn.status != Transaction.Status.canceled:
                ret[txn.currency] = ret.get(txn.currency, 0) + txn.balance_amount()
        return ret

    def _gen_subaddress(self, account_id: str) -> Subaddress:
        sub = self.store.next_id().to_bytes(8, byteorder="big")
        return self.store.create(Subaddress, account_id=account_id, subaddress_hex=sub.hex())

    def _txn_metadata(self, txn: Transaction) -> Tuple[bytes, bytes]:
        if self.offchain.is_under_dual_attestation_limit(txn.currency, txn.amount):
            if txn.refund_diem_txn_version and txn.refund_reason:
                return self.diem_account.refund_metadata(txn.refund_diem_txn_version, txn.refund_reason)  # pyre-ignore
            if txn.subaddress_hex:
                return self.diem_account.general_metadata(txn.subaddress(), str(txn.payee))
        elif txn.reference_id:
            cmd = self.store.find(PaymentCommand, reference_id=txn.reference_id)
            return self.diem_account.travel_metadata(cmd.to_offchain_command())
        raise ValueError("could not create diem payment transacton metadata: %s" % txn)


class OffChainAPI(Base):
    def offchain_api(self, request_sender_address: str, request_bytes: bytes) -> CommandResponseObject:
        cmd = self.offchain.process_inbound_request(request_sender_address, request_bytes)
        getattr(self, "_handle_offchain_%s" % utils.to_snake(cmd))(cmd)
        return offchain.reply_request(cid=cmd.id())

    def jws_serialize(self, resp: CommandResponseObject) -> bytes:
        return offchain.jws.serialize(resp, self.diem_account.account.compliance_key.sign)

    def _handle_offchain_payment_command(self, new_offchain_cmd: offchain.PaymentCommand) -> None:
        try:
            cmd = self.store.find(PaymentCommand, reference_id=new_offchain_cmd.reference_id())
            if new_offchain_cmd != cmd.to_offchain_command():
                self._update_payment_command(cmd, new_offchain_cmd)
        except NotFoundError:
            subaddress = utils.hex(new_offchain_cmd.my_subaddress(self.diem_account.account.hrp))
            account_id = self.store.find(Subaddress, subaddress_hex=subaddress).account_id
            self._create_payment_command(account_id, new_offchain_cmd, validate=True)

    def _create_payment_command(self, account_id: str, cmd: offchain.PaymentCommand, validate: bool = False) -> None:
        self.store.create(
            PaymentCommand,
            before_create=lambda _: cmd.validate(None) if validate else None,
            account_id=account_id,
            is_sender=cmd.is_sender(),
            reference_id=cmd.reference_id(),
            is_inbound=cmd.is_inbound(),
            cid=cmd.id(),
            payment_object=asdict(cmd.payment),
        )

    def _update_payment_command(
        self, cmd: PaymentCommand, offchain_cmd: offchain.PaymentCommand, validate: bool = False
    ) -> None:
        self.store.update(
            cmd,
            before_update=lambda _: offchain_cmd.validate(cmd.to_offchain_command()) if validate else None,
            cid=offchain_cmd.id(),
            is_inbound=offchain_cmd.is_inbound(),
            is_abort=offchain_cmd.is_abort(),
            is_ready=offchain_cmd.is_both_ready(),
            payment_object=asdict(offchain_cmd.payment),
        )


class BackgroundTasks(OffChainAPI):
    def start_sync(self, delay: float = 0.1) -> None:
        try:
            self._process_offchain_commands()
            self._send_pending_payments()
            self.event_puller.fetch(self.event_puller.save_payment_txn)
        except Exception as e:
            self.logger.exception(e)
            # Exponential backoff for the delay to slow down the background tasks execution
            # cycle and reduce the duplicated error logs.
            delay = delay * 2

        if threading.main_thread().is_alive():
            threading.Timer(delay, self.start_sync, args=[delay]).start()

    def _send_pending_payments(self) -> None:
        for txn in self.store.find_all(Transaction, status=Transaction.Status.pending):
            self.logger.info("processing %s" % txn)
            try:
                if self.offchain.is_my_account_id(str(txn.payee)):
                    self._send_internal_payment_txn(txn)
                else:
                    self._send_external_payment_txn(txn)
            except jsonrpc.JsonRpcError as e:
                msg = "ignore error %s when sending transaction %s, retry later"
                self.logger.info(msg % (e, txn), exc_info=True)
            except Exception as e:
                msg = "send pending transaction failed with %s, cancel transaction %s."
                self.logger.error(msg % (e, txn), exc_info=True)
                self.store.update(txn, status=Transaction.Status.canceled, cancel_reason=str(e))

    def _send_internal_payment_txn(self, txn: Transaction) -> None:
        _, payee_subaddress = self.diem_account.account.decode_account_identifier(str(txn.payee))
        subaddress = self.store.find(Subaddress, subaddress_hex=utils.hex(payee_subaddress))
        self.store.create(
            Transaction,
            account_id=subaddress.account_id,
            currency=txn.currency,
            amount=txn.amount,
            status=Transaction.Status.completed,
        )
        self.store.update(txn, status=Transaction.Status.completed)

    def _send_external_payment_txn(self, txn: Transaction) -> None:
        if txn.signed_transaction:
            try:
                diem_txn = self.diem_account.client.wait_for_transaction(str(txn.signed_transaction), timeout_secs=0.1)
                self.store.update(txn, status=Transaction.Status.completed, diem_transaction_version=diem_txn.version)
            except jsonrpc.WaitForTransactionTimeout as e:
                self.store.create_event(txn.account_id, "info", str(e))
            except jsonrpc.TransactionHashMismatchError as e:
                self.store.create_event(txn.account_id, "info", str(e))
                self.store.update(txn, signed_transaction=self.diem_account.submit_p2p(txn, self._txn_metadata(txn)))
            except (jsonrpc.TransactionExpired, jsonrpc.TransactionExecutionFailed) as e:
                self.store.create_event(txn.account_id, "info", str(e))
                reason = "something went wrong with transaction execution: %s" % e
                self.store.update(txn, status=Transaction.Status.canceled, cancel_reason=reason)
        else:
            self._start_external_payment_txn(txn)

    def _start_external_payment_txn(self, txn: Transaction) -> None:
        if not txn.subaddress_hex:
            self.store.update(txn, subaddress_hex=self._gen_subaddress(txn.account_id).subaddress_hex)
        if self.offchain.is_under_dual_attestation_limit(txn.currency, txn.amount):
            if not txn.signed_transaction:
                signed_txn = self.diem_account.submit_p2p(txn, self._txn_metadata(txn))
                self.store.update(txn, signed_transaction=signed_txn)
        else:
            if txn.reference_id:
                cmd = self.store.find(PaymentCommand, reference_id=txn.reference_id)
                if cmd.is_sender:
                    if cmd.is_abort:
                        status = Transaction.Status.canceled
                        self.store.update(txn, status=status, cancel_reason="exchange kyc data abort")
                    elif cmd.is_ready:
                        signed_txn = self.diem_account.submit_p2p(txn, self._txn_metadata(txn))
                        self.store.update(txn, signed_transaction=signed_txn)
            else:
                account = self.store.find(Account, id=txn.account_id)
                command = offchain.PaymentCommand.init(
                    sender_account_id=self.diem_account.account.account_identifier(txn.subaddress()),
                    sender_kyc_data=account.kyc_data_object(),
                    currency=txn.currency,
                    amount=txn.amount,
                    receiver_account_id=str(txn.payee),
                )
                self.offchain.send_command(command, self.diem_account.account.compliance_key.sign)
                self._create_payment_command(txn.account_id, command)
                self.store.update(txn, reference_id=command.reference_id())

    def _process_offchain_commands(self) -> None:
        cmds = self.store.find_all(PaymentCommand, is_inbound=True, is_abort=False, is_ready=False, process_error=None)
        for cmd in cmds:
            try:
                offchain_cmd = cmd.to_offchain_command()
                action = offchain_cmd.follow_up_action()
                if action:
                    fn = getattr(self, "_offchain_action_%s" % action.value)
                    new_offchain_cmd = fn(cmd.account_id, offchain_cmd)
                    self.offchain.send_command(new_offchain_cmd, self.diem_account.account.compliance_key.sign)
                    self._update_payment_command(cmd, new_offchain_cmd)
            except Exception as e:
                self.logger.exception(e)
                self.store.update(cmd, process_error=str(e))

    def _offchain_action_evaluate_kyc_data(self, account_id: str, cmd: offchain.PaymentCommand) -> offchain.Command:
        op_kyc_data = cmd.counterparty_actor_obj().kyc_data
        if op_kyc_data is None or self.kyc_sample.match_kyc_data("reject", op_kyc_data):
            return self._new_reject_kyc_data(cmd, "KYC data is rejected")
        elif self.kyc_sample.match_any_kyc_data(["soft_match", "soft_reject"], op_kyc_data):
            return cmd.new_command(status=Status.soft_match)
        return self._ready_for_settlement(account_id, cmd)

    def _offchain_action_clear_soft_match(self, account_id: str, cmd: offchain.PaymentCommand) -> offchain.Command:
        return cmd.new_command(additional_kyc_data="{%r: %r}" % ("account_id", account_id))

    def _offchain_action_review_kyc_data(self, account_id: str, cmd: offchain.PaymentCommand) -> offchain.Command:
        op_kyc_data = cmd.counterparty_actor_obj().kyc_data
        if op_kyc_data is None or self.kyc_sample.match_kyc_data("soft_reject", op_kyc_data):
            return self._new_reject_kyc_data(cmd, "KYC data review result is reject")
        return self._ready_for_settlement(account_id, cmd)

    def _new_reject_kyc_data(self, cmd: offchain.PaymentCommand, msg: str) -> offchain.Command:
        return cmd.new_command(status=Status.abort, abort_code=AbortCode.reject_kyc_data, abort_message=msg)

    def _ready_for_settlement(self, account_id: str, cmd: offchain.PaymentCommand) -> offchain.Command:
        if cmd.is_sender():
            return cmd.new_command(status=Status.ready_for_settlement)

        sig_msg = cmd.travel_rule_metadata_signature_message(self.diem_account.account.hrp)
        sig = self.diem_account.account.compliance_key.sign(sig_msg).hex()
        kyc_data = self.store.find(Account, id=account_id).kyc_data_object()
        return cmd.new_command(recipient_signature=sig, kyc_data=kyc_data, status=Status.ready_for_settlement)


class App(BackgroundTasks):
    def create_account(self, data: JsonInput) -> Account:
        account = self.store.create(
            Account,
            kyc_data=data.get_nullable("kyc_data", str, self._validate_kyc_data),
        )
        balances = data.get_nullable("balances", dict)
        if balances:
            for c, a in balances.items():
                self._create_transaction(
                    account.id, Transaction.Status.completed, JsonInput({"currency": c, "amount": a})
                )
        return account

    def create_account_payment(self, account_id: str, data: JsonInput) -> Payment:
        self.store.find(Account, id=account_id)
        payee = data.get("payee", str, self._validate_account_identifier)
        txn = self._create_transaction(account_id, Transaction.Status.pending, data, payee=payee)
        return Payment(id=txn.id, account_id=account_id, payee=payee, currency=txn.currency, amount=txn.amount)

    def create_account_payment_uri(self, account_id: str, data: JsonInput) -> PaymentUri:
        currency = data.get_nullable("currency", str, self._validate_currency_code)
        amount = data.get_nullable("amount", int, self._validate_amount)
        self.store.find(Account, id=account_id)
        sub = self._gen_subaddress(account_id)
        diem_acc_id = self.diem_account.account.account_identifier(sub.subaddress_hex)
        uri = identifier.encode_intent(diem_acc_id, currency, amount)
        return PaymentUri(id=sub.id, account_id=account_id, currency=currency, amount=amount, payment_uri=uri)

    def get_account_balances(self, account_id: str) -> Dict[str, int]:
        self.store.find(Account, id=account_id)
        return self._balances(account_id)

    def get_account_events(self, account_id: str) -> List[Event]:
        return self.store.find_all(Event, account_id=account_id)

    def _create_transaction(
        self, account_id: str, status: str, data: JsonInput, payee: Optional[str] = None
    ) -> Transaction:
        return self.store.create(
            Transaction,
            account_id=account_id,
            currency=data.get("currency", str, self._validate_currency_code),
            amount=data.get("amount", int, self._validate_amount),
            payee=payee,
            status=status,
            before_create=self._validate_account_balance,
        )



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass, field, asdict, fields
from enum import Enum
from typing import Optional, List, Dict, Any
from .... import identifier, offchain, diem_types


@dataclass
class Base:
    id: str


@dataclass
class Account(Base):
    kyc_data: Optional[str] = field(default=None)

    def kyc_data_object(self) -> offchain.KycDataObject:
        if self.kyc_data:
            return offchain.from_json(str(self.kyc_data), offchain.KycDataObject)
        else:
            return offchain.individual_kyc_data()


@dataclass
class PaymentUri(Base):
    account_id: str
    payment_uri: str
    currency: Optional[str] = field(default=None)
    amount: Optional[int] = field(default=None)

    def intent(self, hrp: str) -> identifier.Intent:
        return identifier.decode_intent(self.payment_uri, hrp)


@dataclass
class Subaddress(Base):
    account_id: str
    subaddress_hex: str


@dataclass
class Payment(Base):
    account_id: str
    currency: str
    amount: int
    payee: str


@dataclass
class Event(Base):
    account_id: str
    type: str
    data: str
    timestamp: int


@dataclass
class KycSample:
    minimum: str
    reject: str
    soft_match: str
    soft_reject: str

    @staticmethod
    def gen(surname: str) -> "KycSample":
        def gen_kyc_data(name: str) -> str:
            return offchain.to_json(offchain.individual_kyc_data(given_name=name, surname=surname))

        return KycSample(**{f.name: gen_kyc_data("%s-kyc" % f.name) for f in fields(KycSample)})

    def match_kyc_data(self, field: str, kyc: offchain.KycDataObject) -> bool:
        subset = asdict(offchain.from_json(getattr(self, field), offchain.KycDataObject))
        return all(getattr(kyc, k) == v for k, v in subset.items() if v)

    def match_any_kyc_data(self, fields: List[str], kyc: offchain.KycDataObject) -> bool:
        return any(self.match_kyc_data(f, kyc) for f in fields)


class RefundReason(str, Enum):
    invalid_subaddress = "invalid_subaddress"
    other = "other"

    @staticmethod
    def from_diem_type(reason: diem_types.RefundReason) -> "RefundReason":
        if isinstance(reason, diem_types.RefundReason__InvalidSubaddress):
            return RefundReason.invalid_subaddress
        return RefundReason.other

    def to_diem_type(self) -> diem_types.RefundReason:
        return diem_types.RefundReason__InvalidSubaddress()


@dataclass
class Transaction(Base):
    class Status(str, Enum):
        completed = "completed"
        canceled = "canceled"
        pending = "pending"

    account_id: str
    currency: str
    amount: int
    status: Status
    cancel_reason: Optional[str] = field(default=None)
    payee: Optional[str] = field(default=None)
    subaddress_hex: Optional[str] = field(default=None)
    reference_id: Optional[str] = field(default=None)
    signed_transaction: Optional[str] = field(default=None)
    diem_transaction_version: Optional[int] = field(default=None)
    refund_diem_txn_version: Optional[int] = field(default=None)
    refund_reason: Optional[RefundReason] = field(default=None)

    def subaddress(self) -> bytes:
        return bytes.fromhex(str(self.subaddress_hex))

    def balance_amount(self) -> int:
        return -self.amount if self.payee else self.amount


@dataclass
class PaymentCommand(Base):
    account_id: str
    reference_id: str
    cid: str
    is_sender: bool
    payment_object: Dict[str, Any]
    is_inbound: bool = field(default=False)
    is_abort: bool = field(default=False)
    is_ready: bool = field(default=False)
    process_error: Optional[str] = field(default=None)

    def to_offchain_command(self) -> offchain.PaymentCommand:
        payment = offchain.from_dict(self.payment_object, offchain.PaymentObject)
        return offchain.PaymentCommand(
            my_actor_address=payment.sender.address if self.is_sender else payment.receiver.address,
            payment=payment,
            inbound=self.is_inbound,
            cid=self.cid,
        )
